/*
 * Copyright (c) 2020, Derek Buitenhuis
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * This file was mostly autogenerated by 'genstruct', then cleaned up manually after.
 */

#ifndef _OBUPARSE_JSON_INTERNAL
#define _OBUPARSE_JSON_INTERNAL

#include "obuparse.h"

void print_json_film_grain_params(OBPFilmGrainParameters *my_struct);
void print_json_frame_header(OBPFrameHeader *my_struct);
void print_json_sequence_header(OBPSequenceHeader *my_struct);
void print_json_metadata(OBPMetadata *my_struct);
void print_json_tile_list(OBPTileList *my_struct);
void print_json_tile_group(OBPTileGroup *my_struct);

#endif
